/*
 * Test pipe operations.
 *
 * Copyright (C) 2020, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <thread>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h> /* sleep */

#include <fsclient/client.h>
#include <systypes/fcntl.h>
#include <systypes/format.h>

#define FMToffsetd "%" pFMToffset "d"



/* Common size and iteration details. */

const offset_t TO_TRANSFER = 600;
const int NUMBER_OF_ITERATIONS = 3;



/* Use the writer to fill the pipe with data. */

static void write_pipe(file_t *writer, int number)
{
  char buffer[TO_TRANSFER];
  long err;

  /* Make writers blocking to permit synchronisation. */

  if ((err = client_set_blocking(writer, NOTIFY_SPACE_AVAILABLE)))
  {
    printf("Could not set pipe #%d as blocking: %s\n", number, l4sys_errtostr(err));
    return;
  }

  for (int loop = 0; loop < NUMBER_OF_ITERATIONS; loop++)
  {
    for (int region = 0; region < 26; region++)
    {
      memset(buffer, (int) 'a' + region, TO_TRANSFER);

      offset_t nwritten = client_write(writer, buffer, TO_TRANSFER);

      printf("Written " FMToffsetd "/" FMToffsetd " in #%d of %d/%d to pipe #%d...\n", nwritten, TO_TRANSFER, region, loop, 2, number);
    }

    sleep(1);
  }

  /* Flush to make the final output available. */

  client_close(writer);
}

/* Use the reader to obtain data from the pipe. */

static void read_pipes(file_t *reader1, file_t *reader2)
{
  offset_t totals[] = {0, 0};
  unsigned int active = 2;
  long err;
  file_t *reader;

  /* Use a local notifier to wait for pipe events. */

  file_notifier_t *notifier = client_notifier_local();

  /* Register the readers for notification. */

  if ((err = client_subscribe(reader1, NOTIFY_CONTENT_AVAILABLE | NOTIFY_PEER_CLOSED, notifier)) ||
      (err = client_subscribe(reader2, NOTIFY_CONTENT_AVAILABLE | NOTIFY_PEER_CLOSED, notifier)))
  {
    printf("Could not subscribe to notifications: %s\n", l4sys_errtostr(err));
    return;
  }

  while (1)
  {
    char buffer[TO_TRANSFER];
    offset_t nread;

    /* Wait for notification of content. */

    err = client_wait_files(&reader, notifier);

    if (err)
    {
      printf("Error waiting for notifications: %s\n", l4sys_errtostr(err));
      client_notifier_close(notifier);
      return;
    }

    if ((reader != reader1) && (reader != reader2))
    {
      printf("Spurious notification received for %p versus %p, %p.\n", reader, reader1, reader2);
      continue;
    }

    int p = reader == reader1 ? 0 : 1;

    printf("Pipe #%d notified with conditions:%s%s\n", p + 1, reader->notifications & NOTIFY_PEER_CLOSED ? " closed" : "",
                                                              reader->notifications & NOTIFY_CONTENT_AVAILABLE ? " content" : "");

    nread = client_read(reader, buffer, TO_TRANSFER);

    while (nread)
    {
      totals[p] += nread;

      printf("Read " FMToffsetd "/" FMToffsetd ", total " FMToffsetd ", first %c, last %c, from pipe #%d...\n", nread, TO_TRANSFER, totals[p], *buffer, *(buffer + nread - 1), p + 1);
      nread = client_read(reader, buffer, TO_TRANSFER);
    }

    /* Without any more content, a peer closed event should terminate reading
       from the pipe. */

    if (reader->notifications & NOTIFY_PEER_CLOSED)
    {
      active--;
      if (!active)
        break;
    }
  }

  client_close(reader1);
  client_close(reader2);
  client_notifier_close(notifier);

  for (int p = 0; p < 2; p++)
  {
    printf("Data received for pipe #%d: " FMToffsetd "\n", p, totals[p]);
    printf("Data shown for pipe #%d: %s\n", p, totals[p] == TO_TRANSFER * 26 * NUMBER_OF_ITERATIONS ? "True" : "False");
  }
}

int main(void)
{
  /* Obtain pipe endpoint references. Blocking will be set up manually. */

  printf("Obtaining pipes...\n");

  file_t *reader1, *reader2, *writer1, *writer2;
  long err = client_pipe(&reader1, &writer1, O_NONBLOCK) ||
             client_pipe(&reader2, &writer2, O_NONBLOCK);

  if (err)
  {
    printf("Could not obtain pipes: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Starting activities...\n");

  /* Schedule reader and writer threads. */

  std::thread *activities[3];

  activities[0] = new std::thread(read_pipes, reader1, reader2);
  activities[1] = new std::thread(write_pipe, writer1, 1);
  activities[2] = new std::thread(write_pipe, writer2, 2);

  for (int i = 0; i < 3; i++)
    activities[i]->join();

  printf("End of test.\n");
  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
